#include "stdafx.h"
#include "MouseView.h"

static const GLfloat MOUSERESO (0.02);
static const GLfloat ROTRESO (0.1);
static const GLfloat MAX_DISTANCE (10.0f);

CMouseView::CMouseView(void)
{
    m_vPosition.setX(0.0);
    m_vPosition.setY(0.0);
    m_vPosition.setZ(15.0);
    m_vTarget.setX(0.0);
    m_vTarget.setY(0.0);
    m_vTarget.setZ(0.0);
    m_vUp.setX(0.0);
    m_vUp.setY(1.0);
    m_vUp.setZ(0.0);
    update();
    viewport[0] = 0;
    viewport[1] = 0;
    viewport[2] = 640;
    viewport[3] = 480;
    fovy = 60.0;
    aspect = 640.0/480.0;
    zNear = 1.0;
    zFar = 100.0;
    width = 640;
    height = 480;
}

CMouseView::~CMouseView(void)
{
}

// }EX̏XV
void CMouseView::mouse(int button, int state, int x, int y)
{
    switch(button) 
    {
    case GLUT_LEFT_BUTTON:
        if (state == GLUT_DOWN) 
        {
            onButton[0] = TRUE;
            startPos[0][0] = x;
            startPos[0][1] = y;
            backPos[0] = x;
            backPos[1] = y;
        }
        else
        {
            onButton[0] = FALSE;
        }
        break;
    case GLUT_MIDDLE_BUTTON:
        if (state == GLUT_DOWN) 
        {
            onButton[1] = TRUE;
            startPos[1][0] = x;
            startPos[1][1] = y;
            backPos[0] = x;
            backPos[1] = y;
        }
        else 
        {
            onButton[1] = FALSE;
        }
        break;
    case GLUT_RIGHT_BUTTON:
        if (state == GLUT_DOWN)
        {
            onButton[2] = TRUE;
            startPos[2][0] = x;
            startPos[2][1] = y;
            backPos[0] = x;
            backPos[1] = y;
        }
        else
        {
            onButton[2] = FALSE;
        }
    }
}

void CMouseView::motion(int x, int y)
{
    GLfloat diff[2], diffPos[3][2];
    diff[0] = x - backPos[0];
    diff[1] = y - backPos[1];

    if (onButton[2])
    {
        // J̌ύX
        // ̌vZ
        float len = sqrt(diff[0]*diff[0]+diff[1]*diff[1]);
        // 0傫
        if (len > 0.0)
        {
            // J̕xNg
            btVector3 vecX = m_mRotation * btVector3(1, 0, 0);
            btVector3 vecY = m_mRotation * btVector3(0, 1, 0);
            btVector3 vecZ = m_mRotation * btVector3(0, 0, 1);
            // }EX̃hbOxNg
            btVector3 vecDrag = (vecX*diff[0] + vecY*diff[1]).normalize();
            // 2̃xNg]厲xNgvZ
            btVector3 vaxis = vecZ.cross(vecDrag);
            // hbO]pxvZ
            btScalar angle = ToRadian(len*ROTRESO);
            // NH[^jI
            btQuaternion qt;
            qt.setRotation(vaxis, angle);
            // ]s񐶐
            btMatrix3x3 mat;
            mat.setRotation(qt);
            // ݂̉]sɊ|
            m_mRotation = mat * m_mRotation;
            // _̍WXV
            btScalar len2 = (m_vTarget-m_vPosition).length();
            // _ʒuJ̕֌̒ړ
            m_vTarget = m_vPosition + len2 * (m_mRotation * btVector3(0, 0, 1));
        }
    }
    if (onButton[1])
    {
        // sړ
        btVector3 vecX = m_mRotation * btVector3(1, 0, 0);
        btVector3 vecY = m_mRotation * btVector3(0, 1, 0);
        m_vPosition += MOUSERESO*(diff[0]*vecX + diff[1]*vecY);
        m_vTarget   += MOUSERESO*(diff[0]*vecX + diff[1]*vecY);
    }
#if 0
    if (onButton[0])
    {
        // Jړ
        diffPos[2][0] = x - startPos[2][0];
        diffPos[2][1] = y - startPos[2][1];
        btVector3 dir = m_mRotation * btVector3(0, 0, 1);
        m_vPosition += MOUSERESO*(-diff[0] + diff[1])*dir;
        m_vTarget   += MOUSERESO*(-diff[0] + diff[1])*dir;
    }
#endif
    backPos[0] = x;
    backPos[1] = y;

    // ǂꂩ{^Ăꍇ͎_̍XV
    if (onButton[0] || onButton[1] || onButton[2])
    {
        update();
    }
}

void CMouseView::setViewport(int x, int y, int w, int h)
{
    viewport[0] = x;
    viewport[1] = y;
    viewport[2] = w;
    viewport[3] = h;
    aspect = (double)w/(double)h;
    width = w;
    height = h;
}

void CMouseView::setPerspective(GLfloat fov, GLfloat as, GLfloat n, GLfloat f)
{
    fovy = fov;
    aspect = as;
    zNear = n;
    zFar = f;
}

void CMouseView::update()
{
    eye[0] = m_vPosition.getX();
    eye[1] = m_vPosition.getY();
    eye[2] = m_vPosition.getZ();
    center[0] = m_vTarget.getX();
    center[1] = m_vTarget.getY();
    center[2] = m_vTarget.getZ();
    up[0] = 0;
    up[1] = 1;
    up[2] = 0;
    calculateCameraMatrix(m_vPosition, m_vTarget);
}

void CMouseView::calculateCameraMatrix(btVector3 &position, btVector3 &target)
{
    btVector3 xAxis, yAxis, zAxis;
    zAxis = (target - position).normalized();
    // J̌Yǂ
    if (zAxis != btVector3(0, 1, 0))
    {
        yAxis = btVector3(0, 1, 0);
    }
    else
    {
        yAxis = btVector3(0, 0, -1);
    }
    // XZY̊O
    xAxis = yAxis.cross(zAxis);
    xAxis.normalize();
    // YZX̊O
    yAxis = zAxis.cross(xAxis);
    yAxis.normalize();

    // ꂪJ̍Wn
    m_mRotation.setValue(
        xAxis.x(), yAxis.x(), zAxis.x(),
        xAxis.y(), yAxis.y(), zAxis.y(),
        xAxis.z(), yAxis.z(), zAxis.z());
}

void CMouseView::calculatePoint(int x, int y, double pos[3])
{
	// eʂ3Ԃł̍vZ
	realHeight = 2.0 * zNear * tan(ToRadian(fovy)/2.0);
	// eʂ3Ԃł̕vZ
	realWidth = aspect * realHeight;
//    printf("zNear=%f, fovy=%f, realHeight=%f, realWidth=%f, aspect=%f, width=%d, height=%d\n",
//        zNear, fovy, realHeight, realWidth, aspect, width, height);
    GLfloat rightVec[3], upVec[3], targetVec[3];
    for (int i=0; i<3; i++)
    {
        rightVec[i] = (m_mRotation*btVector3(1,0,0))[i];
        upVec[i] = (m_mRotation*btVector3(0,1,0))[i];
        targetVec[i] = (m_mRotation*btVector3(0,0,1))[i];
    }

	// eʂ̒SWJ̈ʒu{OxNg~߃Nbvʂ܂ł̋
	for (int i=0; i<3; i++)
	{
		centerPos[i] = m_vPosition[i] + zNear * targetVec[i];
	}
	// }EX̍WƓeʂ̃TCY烏[hWnł̈ړdx, dy߂
	dx = ((double)x/(double)width - 0.5) * realWidth;
	dy = ((double)y/(double)height - 0.5) * realHeight;
	// ĽvZʂƃJ̍WnNbNW߂
	for (int i=0; i<3; i++)
	{
		clickPos[i] = centerPos[i] + rightVec[i] * (-dx) + upVec[i] * (-dy);
		pos[i] = clickPos[i];
	}
//    printf("rightVec=%f %f %f, up=%f %f %f\n", rightVec[0], rightVec[1], rightVec[2], upVec[0], upVec[1], upVec[2]);
//    printf("dx=%f, dy=%f\n", dx, dy);
//    printf("clickPos=%f %f %f\n", clickPos[0], clickPos[1], clickPos[2]);
    for (int i=0; i<3; i++)
    {
        m_vDir[i] = (btScalar)(clickPos[i] - eye[i]);
    }
    m_vDir.normalize();
//    printf("dir=%f %f %f\n", m_vDir[0], m_vDir[1], m_vDir[2]);
}

int CMouseView::rayNearestPoint(double rayOrg[], double rayDelta[], GLfloat **pos, int point_num, float maxT)
{
    // P(t) = Porg + td
    // P(t)FR@PorgF̌_@dF̕PʃxNg@F0`l܂ł̃p[^iC̒j
    // q^ꂽɍł߂R̓_q'߂
    // ̓xNg𓊉e邾ŗǂ
    // v : Progqւ̃xNg
    double t;
    btVector3 q_dash;
    btVector3 tempPos;
    double min = FLT_MAX, dis = 0;
    int index = -1;

    btVector3 v, tempRayDelta(rayDelta[0], rayDelta[1], rayDelta[2]);

    // WCgJԂ
    for (int i=0; i<point_num; i++)
    {
        for (int j=0; j<3; j++)
        {
            // NbNWWCg̍Wւ̃xNgvZ
            v[j] = pos[i][j] - rayOrg[j];
        }
//        printf("v[%d]=%f %f %f\n", i, v[0], v[1], v[2]);
        // CƃWCgW̓ςvZ
        t = v.dot(tempRayDelta);
//        printf("t=%f\n", t);
        for (int j=0; j<3; j++)
        {
            // Cւ̎ˉe
            q_dash[j] = rayOrg[j] + t * rayDelta[j];
            tempPos[j] = pos[i][j];
        }
        // ˉexNgƍW̋vZ
        dis = q_dash.distance(tempPos);
        if (min > dis)
        {
            min = dis;
            index = i;
        }
//        printf("index[%d]=%f pos=%f %f %f, q'=%f %f %f\n", i, dis, pos[i][0], pos[i][1], pos[i][2], q_dash[0], q_dash[1], q_dash[2]);
    }

//    if (dis > MAX_DISTANCE) index = -1;
//    printf("dis[%d]=%f\n", index, dis);

    return index;
}